import logging
import os
import tomllib


print("Path: ", os.getcwd())

logger = logging.getLogger(__name__)
logger.level = logging.INFO


def get_values(*args: str) -> dict | None:
    args = list(args)
    par = config
    while args:
        key = args.pop(0)
        try:
            par = par[key]
            continue
        except KeyError as e:
            logger.info(e.args[0])
            return None
    return par


def get_values2(*args: str) -> tuple[str, int]:
    args2 = list(args)
    par = config
    while args2:
        key = args2.pop(0)
        try:
            par = par[key]
            continue
        except KeyError as e:
            logger.info(e.args[0])
            return ".".join(list(args)), -1
    return par, 0

try:
    with open("./config.toml", "rb") as f:
        config = tomllib.load(f)
except FileNotFoundError as e:
    logger.warning(e)
    config = dict()



import datetime
import unittest
import inspect
from itertools import count
from src.DLMS_SPODES.types.common_data_types import encode_length
from src.DLMS_SPODES.cosem_interface_classes import ic, collection
from src.DLMS_SPODES.types import cdt, cst, ut, implementations as impl, choices
from src.DLMS_SPODES import relation_to_OBIS, enums
from src.DLMS_SPODES.cosem_interface_classes.collection import Collection



class TestType(unittest.TestCase):
    def test_encode_length(self):
        self.assertEqual(encode_length(1), b'\x01')
        self.assertEqual(encode_length(0x7e), b'\x7e')
        self.assertEqual(encode_length(0x80), b'\x81\x80')
        self.assertEqual(encode_length(0xff), b'\x81\xff')
        self.assertEqual(encode_length(0x100), b'\x82\x01\x00')
        self.assertEqual(encode_length(0x1000), b'\x82\x10\x00')
        self.assertEqual(encode_length(0x10000), b'\x84\x00\x01\x00\x00')
        self.assertEqual(encode_length(0xffffffff), b'\x84\xff\xff\xff\xff')

    def test_exist_attr(self):
        """ Existing attribute 'class_name' in each DLMS class """
        # todo: rewrite, don't work in new API
        # for c_id in ic._COSEM_interface_class_ids:
        #     dlms_class = get_type_from_class(c_id, 0)
        #     self.assertTrue(hasattr(dlms_class, 'NAME'), F'{dlms_class}')

    def test_BitString(self):
        pattern = '101011'
        a = cdt.BitString()
        self.assertEqual(a.encoding, b'\x04\x00', 'default initiation')
        a.set(pattern)
        self.assertEqual(cdt.BitString(pattern), a, 'check set_contents_from')
        self.assertEqual(a.decode(), [1, 0, 1, 0, 1, 1], 'decode to list')

    def test_DateTime(self):
        pattern = datetime.datetime(2020, 1, 1, tzinfo=datetime.timezone.utc)
        self.assertEqual(cdt.DateTime(pattern).decode(), pattern, 'init from datetime and decoding')

    def test_Collection(self):
        collection = Collection()
        collection.add_major(collection.create(class_id=ut.CosemClassId(15), version=cdt.Unsigned(1), logical_name=cst.LogicalName('0.0.40.0.0.255')))
        collection.create(class_id=ut.CosemClassId(8), version=cdt.Unsigned(1), logical_name=cst.LogicalName('0.0.1.0.0.255'))
        collection.create(class_id=ut.CosemClassId(3), version=cdt.Unsigned(0), logical_name=cst.LogicalName('1.0.2.29.0.255'))
        self.assertTrue(len(collection) == 3)
        collection.clear()
        self.assertTrue(len(collection) == 1)
        collection.add_if_missing(class_id=ut.CosemClassId(15), version=cdt.Unsigned(1), logical_name=cst.LogicalName('0.0.40.0.0.255'))
        # self.collection.create(class_id=ut.CosemClassId(15), version=cdt.Unsigned(1), logical_name=cst.LogicalName('0.0.40.0.0.255'))
        self.assertTrue(len(collection) == 1, 'check for not added Association again')

    def test_UnitScaler(self):
        value = cdt.ScalUnitType()
        value.set((10, 10))
        print(value)

    def test_ProfileGeneric(self):
        from src.DLMS_SPODES.cosem_interface_classes.association_ln.ver1 import ObjectListElement
        from src.DLMS_SPODES.types.implementations import structs
        col = Collection()
        col.create(class_id=ut.CosemClassId(15), version=cdt.Unsigned(1), logical_name=cst.LogicalName('0.0.40.0.0.255'))
        col.create(class_id=ut.CosemClassId(8), version=cdt.Unsigned(0), logical_name=cst.LogicalName('0.0.1.0.0.255'))
        col.create(class_id=ut.CosemClassId(3), version=cdt.Unsigned(0), logical_name=cst.LogicalName('1.0.2.29.0.255'))
        col.create(class_id=ut.CosemClassId(3), version=cdt.Unsigned(0), logical_name=cst.LogicalName('1.0.1.29.0.255'))
        col.create(class_id=ut.CosemClassId(3), version=cdt.Unsigned(0), logical_name=cst.LogicalName('1.0.3.29.0.255'))
        col.create(class_id=ut.CosemClassId(3), version=cdt.Unsigned(0), logical_name=cst.LogicalName('1.0.4.29.0.255'))
        profile = col.create(class_id=ut.CosemClassId(7), version=cdt.Unsigned(1), logical_name=cst.LogicalName('1.0.94.7.4.255'))
        profile.collection = col
        profile.set_attr(6, structs.CaptureObjectDefinition().encoding)
        profile.set_attr(3, bytes.fromhex('01 05 02 04 12 00 08 09 06 00 00 01 00 00 ff 0f 02 12 00 00 02 04 12 00 03 09 06 01 00 02 1d 00 ff 0f 03 12 00 00 02 04 12 00 03 09 06 01 00 01 1d 00 ff 0f 03 12 00 00 02 04 12 00 03 09 06 01 00 03 1d 00 ff 0f 03 12 00 00 02 04 12 00 03 09 06 01 00 04 1d 00 ff 0f 03 12 00 00'))
        profile.buffer.selective_access.access_selector.set_contents_from(2)
        profile.set_attr(2, bytes.fromhex('01 01 02 05 09 0c 07 6e 08 1f 07 00 00 ff ff 80 00 00 02 02 0f fd 16 1e 02 02 0f fd 16 1e 02 02 0f fd 16 20 02 02 0f fd 16 20'))
        a = ObjectListElement((3, 0, '1.0.1.29.0.255', None))
        b = col.get_object(a)
        b1 = col.get_object(structs.CaptureObjectDefinition((3, '1.0.1.29.0.255', None, None)))
        # b2 = col.get_object(EntryDescriptor())
        desc1 = profile.get_attr_descriptor(2)
        self.assertEqual(desc1.contents, b'\x00\x07\x01\x00^\x07\x04\xff\x02\x01\x02\x02\x04\x06\x00\x00\x00\x01\x06\x00\x00\x00\x00\x12\x00\x01\x12\x00\x00')
        profile.get_attr(2).selective_access.access_selector.set_contents_from(1)
        desc2 = profile.get_attr_descriptor(2)
        self.assertEqual(desc2.contents, b'\x00\x07\x01\x00^\x07\x04\xff\x02\x01\x01\x02\x04\x02\x04\x12\x00\x01\t\x06\x00\x00\x01\x00\x00\xff\x0f\x02\x12\x00\x00\t\x0c\x07\xe4'
                                         b'\x01\x01\xff\xff\xff\xff\xff\x80\x00\xff\t\x0c\x07\xe4\x01\x02\xff\xff\xff\xff\xff\x80\x00\xff\x01\x00')
        # e = EntryDescriptor()
        # e.set_contents_from((1, 0, 1, 0))

    def test_Association(self):
        ass = collection.AssociationLNVer0('0.0.40.0.1.255')
        print(ass)

    def test_Conformance(self):
        from src.DLMS_SPODES.cosem_interface_classes.association_ln.ver0 import Conformance, XDLMSContextType
        c = Conformance()
        a = int('1011011101111', 2)
        c.set(a)
        self.assertEqual(c.decode(), [1, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        context = XDLMSContextType()
        print(context)

    def test_ImageTransfer(self):
        col = Collection()
        col.create(class_id=ut.CosemClassId(18), version=cdt.Unsigned(0), logical_name=cst.LogicalName('0.0.44.0.0.255'))
        print(col)

    def test_UTF8(self):
        value = cdt.Utf8String()
        value.set('ООО "Курганский приборостроительный завод"')
        print(value)

    def test_Duration(self):
        value = impl.double_long_usingneds.DoubleLongUnsignedSecond()
        value.set(1)
        self.assertEqual(value.report, "0 c")

    def test_ScalerUnitType(self):
        value = cdt.ScalUnitType((1, 4))
        value.unit.set(enums.Unit.CURRENT_AMPERE)
        a = str(value.unit)
        a = str(value.unit)
        print(value.unit == cdt.Unit(enums.Unit.CURRENT_AMPERE))

    def test_Array(self):
        obj = collection.Data("1.1.1.1.1.1")
        obj.set_attr(2, b'\x01\x00')
        # check Type setting in empty Array
        obj.set_attr(2, b'\x01\x01\x11\x02')
        a = obj.value.get_copy([1, 3])
        print(a)

    def test_get_copy(self):
        value = cdt.Unsigned(3)
        copy = value.get_copy(4)
        self.assertNotEqual(value, copy, "check different value")

    def test_Enum(self):
        value = cdt.Unit(4)
        match value:
            case cdt.Unit(4): print("ok")

    def test_integers(self):
        value = impl.integers.Only0(0)
        print(value)

    def test_Structs(self):
        for s in cdt.Structure.__subclasses__():
            print(s)
            self.assertIsInstance(s.NAME, str, "check name")
            for el in s.ELEMENTS:
                self.assertIsInstance(el.NAME, str, "check element type name")
                self.assertTrue(isinstance(el.TYPE, choices.CommonDataTypeChoiceBase) or issubclass(el.TYPE, cdt.CommonDataType), F"check element type is CDT: {s}.{el}")

    def test_RestrictionElement(self):
        from src.DLMS_SPODES.cosem_interface_classes.push_setup.ver2 import RestrictionElement
        self.assertEqual(RestrictionElement().encoding, b'\x02\x02\x16\x00\x00', "empty init")
        self.assertEqual(RestrictionElement((0, None)).encoding, b'\x02\x02\x16\x00\x00', "init by None")
        self.assertEqual(RestrictionElement((1, ("01.01.2000", "02.01.2000"))).encoding, b'\x02\x02\x16\x01\x02\x02\t\x05\x07\xd0\x01\x01\xff\t\x05\x07\xd0\x01\x02\xff', "init by DateRestriction")
        self.assertEqual(RestrictionElement((2, (1, 100000))).encoding, b'\x02\x02\x16\x02\x02\x02\x06\x00\x00\x00\x01\x06\x00\x01\x86\xa0', "init by EntryRestriction")
        self.assertEqual(RestrictionElement(b'\x02\x02\x16\x00\x00').decode(), (0, None), "init from bytes by None")
        self.assertEqual(RestrictionElement(b'\x02\x02\x16\x01\x02\x02\t\x05\x07\xd0\x01\x01\xff\t\x05\x07\xd0\x01\x02\xff').encoding, b'\x02\x02\x16\x01\x02\x02\t\x05\x07\xd0\x01\x01\xff\t\x05\x07\xd0\x01\x02\xff', "init from bytes by DateRestriction")
        self.assertEqual(RestrictionElement(b'\x02\x02\x16\x02\x02\x02\x06\x00\x00\x00\x01\x06\x00\x01\x86\xa0').decode(), (2, (1, 100000)), "init from bytes by EntryRestriction")
        value = RestrictionElement()
        value.restriction_type.set(1)
        print('ok')

    def test_cdt_type_name(self):
        from src.DLMS_SPODES.cosem_interface_classes.push_setup.ver2 import RestrictionElement
        value = RestrictionElement()
        print(cdt.get_type_name(value))
        print(cdt.get_type_name(RestrictionElement))
        value = cdt.VisibleString("hello")
        print(cdt.get_type_name(value))
        value = cst.LogicalName("1.1.1.1.1.255")
        print(cdt.get_type_name(value))
        value = cdt.Unsigned(1)
        print(cdt.get_type_name(value))
        value = impl.integers.Only0()
        print(cdt.get_type_name(value))
        print(cdt.get_type_name(value))

    def test_all_cdt(self):
        from src.DLMS_SPODES.cosem_interface_classes.gprs_modem_setup import QualityOfService
        c = [cdt.CommonDataType]
        count1 = count()
        while len(c) != 0:
            t = c.pop()
            c.extend(t.__subclasses__())
            print(next(count1), t, t.TAG, F"{cdt.get_type_name(t)} {t.NAME}" if (not inspect.isabstract(t) and t != cdt.Structure and t != cdt.Enum) else "abstract")



from ...types import common_data_types as cdt


class CommSpeed(cdt.Enum, elements=tuple(range(10))):
    """ The communication speed supported by the corresponding port. This communication speed can be overridden if the HDLC mode of a devive is entered through a special mode
    of another protocol. """

    def decode(self) -> int:
        """ override enum key to enum value"""
        match self.contents:
            case b'\x00':  return 300
            case b'\x01':  return 600
            case b'\x02':  return 1200
            case b'\x03':  return 2400
            case b'\x04':  return 4800
            case b'\x05':  return 9600
            case b'\x06':  return 19200
            case b'\x07':  return 38400
            case b'\x08':  return 57600
            case b'\x09':  return 115200
            case _ as err: raise ValueError(F'Wrong CommSpeed: {err}, expected {CommSpeed.ELEMENTS.keys()}')

    def from_int(self, value: int) -> bytes:
        """ additional cases with Speed Values"""
        match value:
            case 300:    return b'\x00'
            case 600:    return b'\x01'
            case 1200:   return b'\x02'
            case 2400:   return b'\x03'
            case 4800:   return b'\x04'
            case 9600:   return b'\x05'
            case 19200:  return b'\x06'
            case 38400:  return b'\x07'
            case 57600:  return b'\x08'
            case 115200: return b'\x09'
            case other:  return super(CommSpeed, self).from_int(other)


class RestrictionType(cdt.Enum, elements=(0, 1, 2)):
    """"""


class KeyInfoType(cdt.Enum, elements=(0, 1, 2)):
    """"""


class ProtectionType(cdt.Enum, elements=(0, 1, 2, 3)):
    """"""



from .__class_init__ import *
from ..types import choices
from ..types.implementations import integers


class Register(ic.COSEMInterfaceClasses):
    """ A “Register” object stores a process value or a status value with its associated unit. The register object knows
    the nature of the process value or of the status value. The nature of the value is described by the attribute
    “logical name” using the OBIS identification system. """
    NAME = cn.REGISTER
    CLASS_ID = ClassID.REGISTER
    VERSION = Version.V0
    scaler_unit_not_settable: bool
    A_ELEMENTS = (ic.ICAElement(an.VALUE, choices.register, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement(an.SCALER_UNIT, cdt.ScalUnitType))
    M_ELEMENTS = (
        ic.ICMElement(mn.RESET, integers.Only0),
        ic.ICMElement(mn.NEXT_PERIOD, integers.Only0))

    def characteristics_init(self):
        self._cbs_attr_post_init.update({2: self.__set_value_data_type,
                                         3: self.__set_value_scaler_unit})

        self.scaler_unit_not_settable = False
        """ usability scaler unit flag. if True then it not used"""

    @property
    def value(self) -> choices.RegisterValues:
        return self.get_attr(2)

    @property
    def scaler_unit(self) -> cdt.ScalUnitType:
        return self.get_attr(3)

    @property
    def reset(self) -> integers.Only0:
        return self.get_meth(1)

    def __set_value_data_type(self):
        """ When instead of a “Data” object a “Register” object is used, (with the scaler_unit attribute not used or with scaler = 0, unit = 255) then the data types allowed for
        the value attribute of the “Data” interface class are allowed. """
        match self.value:
            case cdt.Array() | cdt.CompactArray() | cdt.Structure(): self.set_attr(3, cdt.ScalUnitType(b'\x02\x02\x0f\x00\x16\xff'))
            case cdt.Digital() | cdt.Float():                        self.value.SCALER_UNIT = self.scaler_unit
            case _:                                                  """ nothing do it """
        match self.scaler_unit:
            case cdt.ScalUnitType(): self.__set_value_scaler_unit()
            case _:                  """ not necessary set Scaler Unit """

    def __set_value_scaler_unit(self):
        match self.value:
            case cdt.Digital() | cdt.Float() if self.value.SCALER_UNIT is None:  self.value.SCALER_UNIT = self.scaler_unit
            case cdt.Digital() | cdt.Float() if self.value.SCALER_UNIT == self.scaler_unit: """ already set """
            case cdt.Digital() | cdt.Float():                                    raise ValueError(F'Got new scaler: {self.scaler_unit} not order with old {self.value.SCALER_UNIT}')
            case _:                                                              """set only for digital"""

    def __str__(self):
        match self.value, self.scaler_unit:
            case None, None: rep = '? ?'
            case cdt.Digital(), _: rep = self.value.report
            case _: rep = F'{self.value} {self.scaler_unit}'
        return F'{super(Register, self).__str__()} {rep}'



from __future__ import annotations
from struct import pack


class AppVersion:
    __major: int
    __minor: int
    __patch: int | None
    __additional: str
    __match_args__ = ('major', 'minor', 'patch')

    def __init__(self, major: int, minor: int, patch: int = None, additional: str = ''):
        self.__major = major
        self.__minor = minor
        self.__patch = patch
        self.__additional = additional

    @classmethod
    def from_str(cls, value: str):
        match value.split(sep='.', maxsplit=2):
            case (major, minor, patch):
                match major.isdigit(), minor.isdigit(), patch.isdigit():
                    case True, True, True:                        return cls(int(major), int(minor), int(patch))
                    case True, True, False:
                        patch_seq = bytearray(patch, 'utf-8')
                        patch_seq.insert(0, ord('0'))  # if not founded integer when patch default value is 0
                        digit_pos = 0
                        while patch_seq[:digit_pos + 1].isdigit() and digit_pos < len(patch_seq):
                            digit_pos += 1
                        return cls(int(major), int(minor), int(patch_seq[:digit_pos]), patch_seq[digit_pos:].decode())
                    case _:                                      return cls(0, 0, 0)
            case (major, minor):
                match major.isdigit(), minor.isdigit():
                    case True, True:                        return cls(int(major), int(minor))
                    case _:                                      return cls(0, 0)
            case _:                                              return cls(0, 0, 0)

    @property
    def major(self) -> int:
        return self.__major

    @property
    def minor(self) -> int:
        return self.__minor

    @property
    def patch(self) -> int | None:
        return self.__patch

    @property
    def additional(self) -> str:
        return self.__additional

    def __eq__(self, other: AppVersion):
        return hash(self) == hash(other) and self.__additional == other.additional

    def __gt__(self, other: AppVersion):
        if self.__major > other.__major:
            return True
        elif self.__major < other.__major:
            return False
        elif self.__minor > other.__minor:
            return True
        elif self.__minor < other.__minor:
            return False
        if self.__patch is None:
            return False
        elif self.__patch > other.__patch:
            return True
        else:
            return False

    def __ge__(self, other: AppVersion):
        if self.__major > other.__major:
            return True
        elif self.__major < other.__major:
            return False
        elif self.__minor > other.__minor:
            return True
        elif self.__minor < other.__minor:
            return False
        if self.__patch is None:
            return True
        elif self.__patch > other.__patch:
            return True
        elif self.__patch < other.__patch:
            return False
        else:
            return True

    def __str__(self):
        return F'{self.__major}.{self.__minor}{F".{self.__patch}" if self.__patch is not None else ""}{F" <{self.__additional}>" if len(self.__additional) != 0 else ""}'

    def __repr__(self):
        return F"{self.__class__.__name__}({self.__major}, {self.__minor}{F', {self.__patch}' if self.__patch is not None else ''}{', additional='+repr(self.additional) if self.additional else ''})"

    def __hash__(self):
        """ to int6 every 2 byte for major, minor, patch """
        if self.__patch is not None:
            return hash((self.__major, self.__minor, self.__patch))
        else:
            return hash((self.__major, self.__minor))

    def select_nearest(self, variants: list[AppVersion | str]) -> AppVersion | None:
        """ select left nearest from list. If version is absence return None """
        app_versions: list[AppVersion | None] = [self]
        for version in variants:
            match version:
                case AppVersion(): app_versions.append(version)
                case str():        app_versions.append(self.from_str(version))
                case _:            pass
        app_versions.sort()
        app_versions.append(None)
        return app_versions[app_versions.index(self)-1]



FAILED_TO_RECEIVE = "Не удалось получить ответ с устройства за заданный промежуток времени"
OPEN_PORT = "Порт открыт"
INVALID_PDU = "Неверные данные протокола(PDU)"
UPDATE_AVAILABLE = "возможно обновление"
UNKNOWN_APPVERSION_TYPE = "Неизвестная версия преложения"
INITIALIZE_CONNECTION = "Инициализация соединения"



from abc import ABC
from itertools import chain
from typing import TypeAlias, Self
from ..types import cdt, ut, cst
from ..types.implementations import structs


class CommonDataTypeChoiceBase(ut.CHOICE, ABC):
    """ For some attributes of some COSEM interface objects, the data type may be chosen at COSEM object instantiation, in the implementation phase
    of the COSEM server. The Server always shall send back the data type and the value of each attribute, so that together with the logical name, an
    unambiguous interpretation is ensured. The list of possible data types is defined in the “Attribute description” section of a COSEM interface
    class specification. DLMS UA 1000-1 Ed. 13. The tag of a type defined using the CHOICE keyword takes the value of the tag of the type from which
    the chosen data value is taken. ITU-T Rec. X.680 | ISO/IEC 8824-1 """
    TYPE = cdt.CommonDataType

    def __init_subclass__(cls, **kwargs):
        cls.ELEMENTS = dict()
        for t in kwargs["types"]:
            if isinstance(t, dict):  # extended choice
                cls.ELEMENTS[int.from_bytes(tuple(t.values())[0].TAG, "big")] = {k: ut.SequenceElement(v.NAME, v) for k, v in t.items()}
            elif issubclass(t, cdt.CommonDataType):
                cls.ELEMENTS[int.from_bytes(t.TAG, "big")] = ut.SequenceElement(t.NAME, t)
            else:
                raise TypeError(F"got {t.__class__} expected cdt or dict")


class SimpleDataTypeChoice(CommonDataTypeChoiceBase, types=cdt.SimpleDataType.__subclasses__()):
    """All Simple Data Types"""


class ComplexDataTypeChoice(CommonDataTypeChoiceBase, types=cdt.ComplexDataType.__subclasses__()):
    """All Complex Data Types"""


class AccessSelectorsChoice(CommonDataTypeChoiceBase, types=(cdt.NullData, cdt.Array)):
    """All Complex Data Types"""


class CommonDataTypeChoice(CommonDataTypeChoiceBase, types=chain(cdt.SimpleDataType.__subclasses__(), cdt.ComplexDataType.__subclasses__())):
    """Types of Data.value"""


class ExtendedRegisterChoice(CommonDataTypeChoiceBase,
                             types=(cdt.NullData, cdt.BitString, cdt.DoubleLongUnsigned, cdt.OctetString, cdt.VisibleString, cdt.Utf8String, cdt.Unsigned, cdt.LongUnsigned,
                                    cdt.Long64Unsigned)):
    """Types of ExtendedRegister.value"""


class RegisterChoice(CommonDataTypeChoiceBase,
                     types=(cdt.NullData, cdt.BitString, cdt.DoubleLongUnsigned, cdt.OctetString, cdt.VisibleString, cdt.Utf8String, cdt.Unsigned, cdt.LongUnsigned,
                            cdt.Long64Unsigned, cdt.DoubleLong, cdt.Integer, cdt.Long, cdt.Long64, cdt.Enum, cdt.Float32, cdt.Float64)):
    """Types of ExtendedRegister.value"""


class AnyDateTimeChoice(CommonDataTypeChoiceBase, types=(cdt.DateTime, cdt.Date, cdt.Time, {12: cst.OctetStringDateTime, 5: cst.OctetStringDate, 4: cst.OctetStringTime})):
    """Date of the event may contain the date only, the time only or both, encoded as specified in 4.1.6.1."""


simple_dt = SimpleDataTypeChoice()
complex_dt = ComplexDataTypeChoice()
common_dt = CommonDataTypeChoice()
extended_register = ExtendedRegisterChoice()
register = RegisterChoice()
access_selectors: cdt.NullData | cdt.Array = AccessSelectorsChoice()
any_date_time = AnyDateTimeChoice()


ExtendedRegisterValues: TypeAlias = cdt.NullData | cdt.BitString | cdt.DoubleLongUnsigned | cdt.OctetString | cdt.VisibleString | cdt.Utf8String | cdt.Unsigned \
                                    | cdt.LongUnsigned | cdt.Long64Unsigned
RegisterValues: TypeAlias = ExtendedRegisterValues | cdt.DoubleLong | cdt.Integer | cdt.Long | cdt.Long64 | cdt.Enum | cdt.Float32 | cdt.Float64


class RestrictionValue(ut.CHOICE):
    TYPE = (cdt.Structure, cdt.NullData)
    ELEMENTS = {0: ut.SequenceElement("no restriction apply", cdt.NullData),
                1: ut.SequenceElement("restriction by date", structs.RestrictionByDate),
                2: ut.SequenceElement("restriction by entry", structs.RestrictionByEntry)}


restriction_value = RestrictionValue()


class IdentifiedKeyInfoOptions(cdt.Enum, elements=(0, 1)):
    """"""


class KEKId(cdt.Enum, elements=(0,)):
    """"""


class WrappedKeyInfoOptions(cdt.Structure):
    kek_id: KEKId
    key_ciphered_data: cdt.OctetString


class AgreedKeyInfoOptions(cdt.Structure):
    kek_id: cdt.OctetString
    key_ciphered_data: cdt.OctetString


class KeyInfoOptions(ut.CHOICE):
    TYPE = (cdt.Enum, cdt.Structure)
    ELEMENTS = {0: ut.SequenceElement("identified_key", IdentifiedKeyInfoOptions),
                1: ut.SequenceElement("wrapped_key", WrappedKeyInfoOptions),
                2: ut.SequenceElement("agreed_key", AgreedKeyInfoOptions)}


key_info_options = KeyInfoOptions()


class StructureMixin:
    """use for cdt.Structure[Enum, CommonDataType]"""
    ELEMENTS: tuple[cdt.StructElement, cdt.StructElement]
    values: list[cdt.Enum, cdt.CommonDataType]
    __len__: int

    def __init__(self, value: bytes | tuple | list | None | bytearray | Self = None):
        def raise_error(set_value):
            raise RuntimeError(F"not can't be set {set_value} to {self}, available only set complex struct value")
        super(StructureMixin, self).__init__(value)
        self.values[0].register_cb_preset(raise_error)

    def from_sequence(self, sequence: tuple):
        if len(sequence) != len(self):
            raise ValueError(F'Struct {self.__class__.__name__} got length:{len(sequence)}, expected length:{len(self)}')
        self.values.append(self.ELEMENTS[0].TYPE(sequence[0]))
        self.values.append(self.ELEMENTS[1].TYPE(int(self.values[0])))
        self.values[1].set(sequence[1])

    def from_content(self, value: bytes):
        el_value, pdu = cdt.get_instance_and_pdu(self.ELEMENTS[0].TYPE, value)
        self.values.append(el_value)
        el_value, pdu = cdt.get_instance_and_pdu(self.ELEMENTS[1].TYPE.ELEMENTS[int(self.values[0])].TYPE, pdu)
        self.values.append(el_value)





import unittest
from src.DLMS_SPODES.types import cdt, cst, ut
from src.DLMS_SPODES.cosem_interface_classes import collection
from src.DLMS_SPODES.relation_to_OBIS import get_name
from src.DLMS_SPODES.configure import get_attr_index


class TestType(unittest.TestCase):

    def test_configure(self):
        ln = cst.LogicalName("1.0.94.7.4.255")
        obj = collection.ProfileGenericVer1(ln)
        print(get_attr_index(obj))
